from typing import Iterable


class IndexerDict(dict):
    # TODO: implement union, for instance.
    # current slice is slice(2,7) and in `update` slice is slice(6, 15) result should be slice(2, 15)
    # for the given coordinate/dimension
    def update(self, list_of_dict):
        if not isinstance(list_of_dict, list):
            list_of_dict = [list_of_dict]
        for ld in list_of_dict:
            for k, v in ld.items():
                if isinstance(v, Iterable) and len(v) == 0:
                    continue
                if self.__contains__(k):
                    super().__setitem__(k, [self.__getitem__(k)])
                    if v not in self[k]:
                        self[k].append(v)
                else:
                    super().__setitem__(k, v)



class AggMixin:
    def min(self):
        # TODO: maybe check if attribute defined
        return self._variable.min()

    def max(self):
        # TODO: maybe check if attribute defined
        return self._variable.max()

    def first(self):
        return self._variable[0]

    def first(self):
        return self._variable[-1]



import json

import numpy as np
import dask.array as da


class GeokubeDetailsJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.float32):
            return float(obj)
        if isinstance(obj, np.datetime64):
            return str(obj)
        return json.JSONEncoder.default(self, obj)


def maybe_convert_to_json_serializable(obj):
    if isinstance(obj, np.ndarray):
        if np.issubdtype(obj.dtype, np.float32) or np.issubdtype(
            obj.dtype, np.float64
        ):
            return obj.astype(float).tolist()
        elif np.issubdtype(obj.dtype, np.int32) or np.issubdtype(
            obj.dtype, np.int64
        ):
            return obj.astype(int).tolist()
        elif np.issubdtype(obj.dtype, np.datetime64):
            return obj.astype(str).tolist()
        else:
            return obj.tolist()
    elif isinstance(obj, da.Array):
        return maybe_convert_to_json_serializable(np.array(obj))
    elif isinstance(obj, dict):
        return {
            k: maybe_convert_to_json_serializable(v) for k, v in obj.items()
        }
    elif isinstance(obj, np.float32) or isinstance(obj, np.float64):
        return float(obj)
    elif isinstance(obj, np.int32) or isinstance(obj, np.int64):
        return int(obj)
    elif isinstance(obj, np.datetime64):
        return str(obj)
    else:
        return obj



import re
from contextlib import suppress
from enum import Enum, unique
from typing import Optional, Union

import dask.array as da
import numpy as np


class MethodType(Enum):
    # From https://cfconventions.org/Data/cf-conventions/cf-conventions-1.7/build/ape.html
    POINT = ("point", "point", [None])
    MAX = ("max", "maximum", [da.nanmax, np.nanmax])
    MIN = ("min", "minimum", [da.nanmin, np.nanmin])
    MEAN = ("mean", "mean", [da.nanmean, np.nanmean])
    SUM = ("sum", "sum", [da.nansum, np.nansum])
    VARIANCE = ("var", "variance", [da.nanvar, np.nanvar])
    STD_DEV = ("std", "standard_deviation", [da.nanstd, np.nanstd])

    # TODO: deal with all: https://cfconventions.org/Data/cf-conventions/cf-conventions-1.7/build/ape.html
    UNDEFINED = ("<undefined>", "<undefined>", [None])

    def __str__(self) -> str:
        return self.value[1]

    @classmethod
    def _missing_(cls, key):
        for _, member in MethodType.__members__.items():
            # NOTE: we check either short name or operator label
            # Label is required to handle cell_methods properly
            if (member.value[0] == key) or (member.value[1] == key):
                return member
        return cls.UNDEFINED

    @property
    def dask_operator(self):
        return self.value[2][0]

    @property
    def numpy_operator(self):
        return self.value[2][-1]


@unique
class RegridMethod(Enum):
    BILINEAR = "bilinear"
    CONSERVATIVE = "conservative"
    CONSERVATIVE_NORMED = "conservative_normed"
    NEAREST_D2S = "nearest_d2s"
    NEAREST_S2D = "nearest_s2d"
    PATCH = "patch"


class LongitudeConvention(Enum):
    POSITIVE_WEST = 1  # 0 to 360
    NEGATIVE_WEST = 2  # -180 to 180


class LatitudeConvention(Enum):
    POSITIVE_TOP = 1  # 90 to -90
    NEGATIVE_TOP = 2  # -90 to 90



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

# This is from Iris framework https://github.com/...
"""
Definitions of coordinate systems.

"""
import importlib
import inspect
import warnings
from abc import ABCMeta, abstractmethod
from typing import Optional

import cartopy.crs as ccrs
import numpy as np
import xarray as xr

from ..utils.serialization import maybe_convert_to_json_serializable

TOL = 1e-5
PREFIX = "crs_"

def add_prefix(txt: str) -> str:
    if txt.startswith(PREFIX):
        return txt
    return f"{PREFIX}{txt}"

def trim_prefix(txt: str) -> str:
    if not txt.startswith(PREFIX):
        return txt
    return txt.removeprefix(PREFIX)

def _arg_default(value, default, cast_as=float):
    """Apply a default value and type for an optional kwarg."""
    if value is None:
        value = default
    value = cast_as(value)
    return value


def _1or2_parallels(arg):
    """Accept 1 or 2 inputs as a tuple of 1 or 2 floats."""
    try:
        values_tuple = tuple(arg)
    except TypeError:
        values_tuple = (arg,)
    values_tuple = tuple([float(x) for x in values_tuple])
    nvals = len(values_tuple)
    if nvals not in (1, 2):
        emsg = "Allows only 1 or 2 parallels or secant latitudes : got {!r}"
        raise ValueError(emsg.format(arg))
    return values_tuple


def _float_or_None(arg):
    """Cast as float, except for allowing None as a distinct valid value."""
    if arg is not None:
        arg = float(arg)
    return arg


class CoordSystem(metaclass=ABCMeta):
    """
    Abstract base class for coordinate systems.

    """

    grid_mapping_name = None

    def __eq__(self, other):
        return (
            issubclass(self.__class__, other.__class__)
            or issubclass(other.__class__, self.__class__)
        ) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        # Must supply __ne__, Python does not defer to __eq__ for
        # negative equality.
        return not (self == other)

    def as_crs_attributes(self):
        if hasattr(self, "_pretty_attrs"):
            return dict(self._pretty_attrs())
        return {k: v for k, v in self.__dict__.items() if v is not None}

    def xml_element(self, doc, attrs=None):
        """Default behaviour for coord systems."""
        # attrs - optional list of (k,v) items, used for alternate output

        xml_element_name = type(self).__name__
        # lower case the first char
        first_char = xml_element_name[0]
        xml_element_name = xml_element_name.replace(
            first_char, first_char.lower(), 1
        )

        coord_system_xml_element = doc.createElement(xml_element_name)

        if attrs is None:
            attrs = self.__dict__.items()
        attrs = sorted(attrs, key=lambda attr: attr[0])

        for name, value in attrs:
            if isinstance(value, float):
                value_str = "{:.16}".format(value)
            elif isinstance(value, np.float32):
                value_str = "{:.8}".format(value)
            else:
                value_str = "{}".format(value)
            coord_system_xml_element.setAttribute(name, value_str)

        return coord_system_xml_element

    @staticmethod
    def _ellipsoid_to_globe(ellipsoid, globe_default):
        if ellipsoid is not None:
            globe = ellipsoid.as_cartopy_globe()
        else:
            globe = globe_default

        return globe

    def to_dict(self):
        return {
            "name": self.grid_mapping_name,
            **maybe_convert_to_json_serializable(self.__dict__),
        }

    @abstractmethod
    def as_cartopy_crs(self):
        """
        Return a cartopy CRS representing our native coordinate
        system.

        """
        pass

    @abstractmethod
    def as_cartopy_projection(self):
        """
        Return a cartopy projection representing our native map.

        This will be the same as the :func:`~CoordSystem.as_cartopy_crs` for
        map projections but for spherical coord systems (which are not map
        projections) we use a map projection, such as PlateCarree.

        """
        pass


class GeogCS(CoordSystem):
    """
    A geographic (ellipsoidal) coordinate system, defined by the shape of
    the Earth and a prime meridian.

    """

    grid_mapping_name = "latitude_longitude"

    def __init__(
        self,
        semi_major_axis=None,
        semi_minor_axis=None,
        inverse_flattening=None,
        longitude_of_prime_meridian=None,
        *args,
        **kwargs
    ):
        """
        Creates a new GeogCS.

        Kwargs:

        * semi_major_axis, semi_minor_axis:
            Axes of ellipsoid, in metres.  At least one must be given
            (see note below).

        * inverse_flattening:
            Can be omitted if both axes given (see note below).
            Defaults to 0.0 .

        * longitude_of_prime_meridian:
            Specifies the prime meridian on the ellipsoid, in degrees.
            Defaults to 0.0 .

        If just semi_major_axis is set, with no semi_minor_axis or
        inverse_flattening, then a perfect sphere is created from the given
        radius.

        If just two of semi_major_axis, semi_minor_axis, and
        inverse_flattening are given the missing element is calculated from the
        formula:
        :math:`flattening = (major - minor) / major`

        Currently, Iris will not allow over-specification (all three ellipsoid
        parameters).

        Examples::

            cs = GeogCS(6371229)
            pp_cs = GeogCS(iris.fileformats.pp.EARTH_RADIUS)
            airy1830 = GeogCS(semi_major_axis=6377563.396,
                              semi_minor_axis=6356256.909)
            airy1830 = GeogCS(semi_major_axis=6377563.396,
                              inverse_flattening=299.3249646)
            custom_cs = GeogCS(6400000, 6300000)

        """
        # No ellipsoid specified? (0 0 0)
        if (
            (semi_major_axis is None)
            and (semi_minor_axis is None)
            and (inverse_flattening is None)
        ):
            raise ValueError("No ellipsoid specified")

        # Perfect sphere (semi_major_axis only)? (1 0 0)
        elif semi_major_axis is not None and (
            semi_minor_axis is None and inverse_flattening is None
        ):
            semi_minor_axis = semi_major_axis
            inverse_flattening = 0.0

        # Calculate semi_major_axis? (0 1 1)
        elif semi_minor_axis is not None and inverse_flattening is not None:
            semi_major_axis_ = -semi_minor_axis / (
                (1.0 - inverse_flattening) / inverse_flattening
            )
            if (
                semi_major_axis is not None
                and abs(semi_major_axis_ - semi_major_axis) > TOL
            ):
                raise ValueError("Ellipsoid is overspecified")
            semi_major_axis = semi_major_axis_

        # Calculate semi_minor_axis? (1 0 1)
        elif semi_major_axis is not None and inverse_flattening is not None:
            semi_minor_axis_ = semi_major_axis - (
                (1.0 / inverse_flattening) * semi_major_axis
            )
            if (
                semi_minor_axis is not None
                and abs(semi_minor_axis_ - semi_minor_axis) > TOL
            ):
                raise ValueError("Ellipsoid is overspecified")
            semi_minor_axis = semi_minor_axis_

        # Calculate inverse_flattening? (1 1 0)
        elif inverse_flattening is None and (
            semi_major_axis is not None and semi_minor_axis is not None
        ):
            if semi_major_axis == semi_minor_axis:
                inverse_flattening = 0.0
            else:
                inverse_flattening = 1.0 / (
                    (semi_major_axis - semi_minor_axis) / semi_major_axis
                )

        # We didn't get enough to specify an ellipse.
        else:
            raise ValueError("Insufficient ellipsoid specification")

        #: Major radius of the ellipsoid in metres.
        self.semi_major_axis = float(semi_major_axis)

        #: Minor radius of the ellipsoid in metres.
        self.semi_minor_axis = float(semi_minor_axis)

        #: :math:`1/f` where :math:`f = (a-b)/a`.
        self.inverse_flattening = float(inverse_flattening)

        #: Describes 'zero' on the ellipsoid in degrees.
        self.longitude_of_prime_meridian = _arg_default(
            longitude_of_prime_meridian, 0
        )

    def _pretty_attrs(self):
        attrs = [("semi_major_axis", self.semi_major_axis)]
        if self.semi_major_axis != self.semi_minor_axis:
            attrs.append(("semi_minor_axis", self.semi_minor_axis))
        if self.longitude_of_prime_meridian != 0.0:
            attrs.append(
                (
                    "longitude_of_prime_meridian",
                    self.longitude_of_prime_meridian,
                )
            )
        return attrs

    def __repr__(self):
        attrs = self._pretty_attrs()
        # Special case for 1 pretty attr
        if len(attrs) == 1 and attrs[0][0] == "semi_major_axis":
            return "GeogCS(%r)" % self.semi_major_axis
        else:
            return "GeogCS(%s)" % ", ".join(
                ["%s=%r" % (k, v) for k, v in attrs]
            )

    def __str__(self):
        attrs = self._pretty_attrs()
        # Special case for 1 pretty attr
        if len(attrs) == 1 and attrs[0][0] == "semi_major_axis":
            return "GeogCS({:.16})".format(self.semi_major_axis)
        else:
            text_attrs = []
            for k, v in attrs:
                if isinstance(v, float):
                    text_attrs.append("{}={:.16}".format(k, v))
                elif isinstance(v, np.float32):
                    text_attrs.append("{}={:.8}".format(k, v))
                else:
                    text_attrs.append("{}={}".format(k, v))
            return "GeogCS({})".format(", ".join(text_attrs))

    def xml_element(self, doc):
        # Special output for spheres
        attrs = self._pretty_attrs()
        if len(attrs) == 1 and attrs[0][0] == "semi_major_axis":
            attrs = [("earth_radius", self.semi_major_axis)]

        return CoordSystem.xml_element(self, doc, attrs)

    def as_cartopy_crs(self):
        return ccrs.Geodetic(self.as_cartopy_globe())

    def as_cartopy_projection(self):
        return ccrs.PlateCarree()

    def as_cartopy_globe(self):
        # Explicitly set `ellipse` to None as a workaround for
        # Cartopy setting WGS84 as the default.
        return ccrs.Globe(
            semimajor_axis=self.semi_major_axis,
            semiminor_axis=self.semi_minor_axis,
            ellipse=None,
        )


class RegularLatLon(GeogCS):
    grid_mapping_name = "latitude_longitude"

    def __init__(self, *args, **kwargs):
        if "semi_major_axis" not in kwargs:
            super().__init__(semi_major_axis=6371229, *args, **kwargs)
        else:
            super().__init__(*args, **kwargs)


class RotatedGeogCS(CoordSystem):
    """
    A coordinate system with rotated pole, on an optional :class:`GeogCS`.

    """

    grid_mapping_name = "rotated_latitude_longitude"

    def __init__(
        self,
        grid_north_pole_latitude,
        grid_north_pole_longitude,
        north_pole_grid_longitude=None,
        ellipsoid=None,
        *args,
        **kwargs
    ):
        """
        Constructs a coordinate system with rotated pole, on an
        optional :class:`GeogCS`.

        Args:

        * grid_north_pole_latitude:
            The true latitude of the rotated pole in degrees.

        * grid_north_pole_longitude:
            The true longitude of the rotated pole in degrees.

        Kwargs:

        * north_pole_grid_longitude:
            Longitude of true north pole in rotated grid, in degrees.
            Defaults to 0.0 .

        * ellipsoid (:class:`GeogCS`):
            If given, defines the ellipsoid.

        Examples::

            rotated_cs = RotatedGeogCS(30, 30)
            another_cs = RotatedGeogCS(30, 30,
                                       ellipsoid=GeogCS(6400000, 6300000))

        """
        #: The true latitude of the rotated pole in degrees.
        self.grid_north_pole_latitude = float(grid_north_pole_latitude)

        #: The true longitude of the rotated pole in degrees.
        self.grid_north_pole_longitude = float(grid_north_pole_longitude)

        #: Longitude of true north pole in rotated grid in degrees.
        self.north_pole_grid_longitude = _arg_default(
            north_pole_grid_longitude, 0
        )

        #: Ellipsoid definition (:class:`GeogCS` or None).
        self.ellipsoid = ellipsoid

    def _pretty_attrs(self):
        attrs = [
            ("grid_north_pole_latitude", self.grid_north_pole_latitude),
            ("grid_north_pole_longitude", self.grid_north_pole_longitude),
        ]
        if self.north_pole_grid_longitude != 0.0:
            attrs.append(
                ("north_pole_grid_longitude", self.north_pole_grid_longitude)
            )
        if self.ellipsoid is not None:
            attrs.append(("ellipsoid", self.ellipsoid))
        return attrs

    def as_crs_attributes(self):
        return dict(self._pretty_attrs())

    def __repr__(self):
        attrs = self._pretty_attrs()
        result = "RotatedGeogCS(%s)" % ", ".join(
            ["%s=%r" % (k, v) for k, v in attrs]
        )
        # Extra prettiness
        result = result.replace("grid_north_pole_latitude=", "")
        result = result.replace("grid_north_pole_longitude=", "")
        return result

    def __str__(self):
        attrs = self._pretty_attrs()
        text_attrs = []
        for k, v in attrs:
            if isinstance(v, float):
                text_attrs.append("{}={:.16}".format(k, v))
            elif isinstance(v, np.float32):
                text_attrs.append("{}={:.8}".format(k, v))
            else:
                text_attrs.append("{}={}".format(k, v))
        result = "RotatedGeogCS({})".format(", ".join(text_attrs))
        # Extra prettiness
        result = result.replace("grid_north_pole_latitude=", "")
        result = result.replace("grid_north_pole_longitude=", "")
        return result

    def xml_element(self, doc):
        return CoordSystem.xml_element(self, doc, self._pretty_attrs())

    def _ccrs_kwargs(self):
        globe = self._ellipsoid_to_globe(self.ellipsoid, None)
        cartopy_kwargs = {
            "central_rotated_longitude": self.north_pole_grid_longitude,
            "pole_longitude": self.grid_north_pole_longitude,
            "pole_latitude": self.grid_north_pole_latitude,
            "globe": globe,
        }

        return cartopy_kwargs

    def as_cartopy_crs(self):
        return ccrs.RotatedGeodetic(**self._ccrs_kwargs())

    def as_cartopy_projection(self):
        return ccrs.RotatedPole(**self._ccrs_kwargs())


class TransverseMercator(CoordSystem):
    """
    A cylindrical map projection, with XY coordinates measured in metres.

    """

    grid_mapping_name = "transverse_mercator"

    def __init__(
        self,
        latitude_of_projection_origin,
        longitude_of_central_meridian,
        false_easting=None,
        false_northing=None,
        scale_factor_at_central_meridian=None,
        ellipsoid=None,
        *args,
        **kwargs
    ):
        """
        Constructs a TransverseMercator object.

        Args:

        * latitude_of_projection_origin:
                True latitude of planar origin in degrees.

        * longitude_of_central_meridian:
                True longitude of planar origin in degrees.

        Kwargs:

        * false_easting:
                X offset from planar origin in metres.
                Defaults to 0.0 .

        * false_northing:
                Y offset from planar origin in metres.
                Defaults to 0.0 .

        * scale_factor_at_central_meridian:
                Reduces the cylinder to slice through the ellipsoid
                (secant form). Used to provide TWO longitudes of zero
                distortion in the area of interest.
                Defaults to 1.0 .

        * ellipsoid (:class:`GeogCS`):
            If given, defines the ellipsoid.

        Example::

            airy1830 = GeogCS(6377563.396, 6356256.909)
            osgb = TransverseMercator(49, -2, 400000, -100000, 0.9996012717,
                                      ellipsoid=airy1830)

        """
        #: True latitude of planar origin in degrees.
        self.latitude_of_projection_origin = float(
            latitude_of_projection_origin
        )

        #: True longitude of planar origin in degrees.
        self.longitude_of_central_meridian = float(
            longitude_of_central_meridian
        )

        #: X offset from planar origin in metres.
        self.false_easting = _arg_default(false_easting, 0)

        #: Y offset from planar origin in metres.
        self.false_northing = _arg_default(false_northing, 0)

        #: Scale factor at the centre longitude.
        self.scale_factor_at_central_meridian = _arg_default(
            scale_factor_at_central_meridian, 1.0
        )

        #: Ellipsoid definition (:class:`GeogCS` or None).
        self.ellipsoid = ellipsoid

    def __repr__(self):
        return (
            "TransverseMercator(latitude_of_projection_origin={!r}, "
            "longitude_of_central_meridian={!r}, false_easting={!r}, "
            "false_northing={!r}, scale_factor_at_central_meridian={!r}, "
            "ellipsoid={!r})".format(
                self.latitude_of_projection_origin,
                self.longitude_of_central_meridian,
                self.false_easting,
                self.false_northing,
                self.scale_factor_at_central_meridian,
                self.ellipsoid,
            )
        )

    def as_cartopy_crs(self):
        globe = self._ellipsoid_to_globe(self.ellipsoid, None)

        return ccrs.TransverseMercator(
            central_longitude=self.longitude_of_central_meridian,
            central_latitude=self.latitude_of_projection_origin,
            false_easting=self.false_easting,
            false_northing=self.false_northing,
            scale_factor=self.scale_factor_at_central_meridian,
            globe=globe,
        )

    def as_cartopy_projection(self):
        return self.as_cartopy_crs()


# TODO: has the same grid_mapping_name. Problem with constructing a CRS!
# class OSGB(TransverseMercator):
#     """A Specific transverse mercator projection on a specific ellipsoid."""

#     def __init__(self):
#         TransverseMercator.__init__(
#             self,
#             49,
#             -2,
#             400000,
#             -100000,
#             0.9996012717,
#             GeogCS(6377563.396, 6356256.909),
#         )

#     def as_cartopy_crs(self):
#         return ccrs.OSGB()

#     def as_cartopy_projection(self):
#         return ccrs.OSGB()

class WebMercator(CoordSystem):
    """A Specific transverse mercator projection on a specific ellipsoid."""

    def as_cartopy_crs(self):
        return ccrs.epsg('3857')

    def as_cartopy_projection(self):
        return ccrs.epsg('3857')

class Orthographic(CoordSystem):
    """
    An orthographic map projection.

    """

    grid_mapping_name = "orthographic"

    def __init__(
        self,
        latitude_of_projection_origin,
        longitude_of_projection_origin,
        false_easting=None,
        false_northing=None,
        ellipsoid=None,
        *args,
        **kwargs
    ):
        """
        Constructs an Orthographic coord system.

        Args:

        * latitude_of_projection_origin:
            True latitude of planar origin in degrees.

        * longitude_of_projection_origin:
            True longitude of planar origin in degrees.

        Kwargs:

        * false_easting:
            X offset from planar origin in metres. Defaults to 0.0 .

        * false_northing:
            Y offset from planar origin in metres. Defaults to 0.0 .

        * ellipsoid (:class:`GeogCS`):
            If given, defines the ellipsoid.

        """
        #: True latitude of planar origin in degrees.
        self.latitude_of_projection_origin = float(
            latitude_of_projection_origin
        )

        #: True longitude of planar origin in degrees.
        self.longitude_of_projection_origin = float(
            longitude_of_projection_origin
        )

        #: X offset from planar origin in metres.
        self.false_easting = _arg_default(false_easting, 0)

        #: Y offset from planar origin in metres.
        self.false_northing = _arg_default(false_northing, 0)

        #: Ellipsoid definition (:class:`GeogCS` or None).
        self.ellipsoid = ellipsoid

    def __repr__(self):
        return (
            "Orthographic(latitude_of_projection_origin={!r}, "
            "longitude_of_projection_origin={!r}, "
            "false_easting={!r}, false_northing={!r}, "
            "ellipsoid={!r})".format(
                self.latitude_of_projection_origin,
                self.longitude_of_projection_origin,
                self.false_easting,
                self.false_northing,
                self.ellipsoid,
            )
        )

    def as_cartopy_crs(self):
        globe = self._ellipsoid_to_globe(self.ellipsoid, ccrs.Globe())

        warnings.warn(
            "Discarding false_easting and false_northing that are "
            "not used by Cartopy."
        )

        return ccrs.Orthographic(
            central_longitude=self.longitude_of_projection_origin,
            central_latitude=self.latitude_of_projection_origin,
            globe=globe,
        )

    def as_cartopy_projection(self):
        return self.as_cartopy_crs()


class VerticalPerspective(CoordSystem):
    """
    A vertical/near-side perspective satellite image map projection.

    """

    grid_mapping_name = "vertical_perspective"

    def __init__(
        self,
        latitude_of_projection_origin,
        longitude_of_projection_origin,
        perspective_point_height,
        false_easting=None,
        false_northing=None,
        ellipsoid=None,
        *args,
        **kwargs
    ):
        """
        Constructs a Vertical Perspective coord system.

        Args:

        * latitude_of_projection_origin:
            True latitude of planar origin in degrees.

        * longitude_of_projection_origin:
            True longitude of planar origin in degrees.

        * perspective_point_height:
            Altitude of satellite in metres above the surface of the
            ellipsoid.

        Kwargs:

        * false_easting:
            X offset from planar origin in metres. Defaults to 0.0 .

        * false_northing:
            Y offset from planar origin in metres. Defaults to 0.0 .

        * ellipsoid (:class:`GeogCS`):
            If given, defines the ellipsoid.

        """
        #: True latitude of planar origin in degrees.
        self.latitude_of_projection_origin = float(
            latitude_of_projection_origin
        )

        #: True longitude of planar origin in degrees.
        self.longitude_of_projection_origin = float(
            longitude_of_projection_origin
        )

        #: Altitude of satellite in metres.
        self.perspective_point_height = float(perspective_point_height)
        # TODO: test if may be cast to float for proj.4

        #: X offset from planar origin in metres.
        self.false_easting = _arg_default(false_easting, 0)

        #: Y offset from planar origin in metres.
        self.false_northing = _arg_default(false_northing, 0)

        #: Ellipsoid definition (:class:`GeogCS` or None).
        self.ellipsoid = ellipsoid

    def __repr__(self):
        return (
            "Vertical Perspective(latitude_of_projection_origin={!r}, "
            "longitude_of_projection_origin={!r}, "
            "perspective_point_height={!r}, "
            "false_easting={!r}, false_northing={!r}, "
            "ellipsoid={!r})".format(
                self.latitude_of_projection_origin,
                self.longitude_of_projection_origin,
                self.perspective_point_height,
                self.false_easting,
                self.false_northing,
                self.ellipsoid,
            )
        )

    def as_cartopy_crs(self):
        globe = self._ellipsoid_to_globe(self.ellipsoid, ccrs.Globe())

        return ccrs.NearsidePerspective(
            central_latitude=self.latitude_of_projection_origin,
            central_longitude=self.longitude_of_projection_origin,
            satellite_height=self.perspective_point_height,
            false_easting=self.false_easting,
            false_northing=self.false_northing,
            globe=globe,
        )

    def as_cartopy_projection(self):
        return self.as_cartopy_crs()


class Geostationary(CoordSystem):
    """
    A geostationary satellite image map projection.

    """

    grid_mapping_name = "geostationary"

    def __init__(
        self,
        latitude_of_projection_origin,
        longitude_of_projection_origin,
        perspective_point_height,
        sweep_angle_axis,
        false_easting=None,
        false_northing=None,
        ellipsoid=None,
        *args,
        **kwargs
    ):
        """
        Constructs a Geostationary coord system.

        Args:

        * latitude_of_projection_origin:
            True latitude of planar origin in degrees.

        * longitude_of_projection_origin:
            True longitude of planar origin in degrees.

        * perspective_point_height:
            Altitude of satellite in metres above the surface of the ellipsoid.

        * sweep_angle_axis (string):
            The axis along which the satellite instrument sweeps - 'x' or 'y'.

        Kwargs:

        * false_easting:
            X offset from planar origin in metres. Defaults to 0.0 .

        * false_northing:
            Y offset from planar origin in metres. Defaults to 0.0 .

        * ellipsoid (:class:`GeogCS`):
            If given, defines the ellipsoid.

        """
        #: True latitude of planar origin in degrees.
        self.latitude_of_projection_origin = float(
            latitude_of_projection_origin
        )
        if self.latitude_of_projection_origin != 0.0:
            raise ValueError(
                "Non-zero latitude of projection currently not"
                " supported by Cartopy."
            )

        #: True longitude of planar origin in degrees.
        self.longitude_of_projection_origin = float(
            longitude_of_projection_origin
        )

        #: Altitude of satellite in metres.
        self.perspective_point_height = float(perspective_point_height)
        # TODO: test if may be cast to float for proj.4

        #: X offset from planar origin in metres.
        self.false_easting = _arg_default(false_easting, 0)

        #: Y offset from planar origin in metres.
        self.false_northing = _arg_default(false_northing, 0)

        #: The sweep angle axis (string 'x' or 'y').
        self.sweep_angle_axis = sweep_angle_axis
        if self.sweep_angle_axis not in ("x", "y"):
            raise ValueError('Invalid sweep_angle_axis - must be "x" or "y"')

        #: Ellipsoid definition (:class:`GeogCS` or None).
        self.ellipsoid = ellipsoid

    def __repr__(self):
        return (
            "Geostationary(latitude_of_projection_origin={!r}, "
            "longitude_of_projection_origin={!r}, "
            "perspective_point_height={!r}, false_easting={!r}, "
            "false_northing={!r}, sweep_angle_axis={!r}, "
            "ellipsoid={!r}".format(
                self.latitude_of_projection_origin,
                self.longitude_of_projection_origin,
                self.perspective_point_height,
                self.false_easting,
                self.false_northing,
                self.sweep_angle_axis,
                self.ellipsoid,
            )
        )

    def as_cartopy_crs(self):
        globe = self._ellipsoid_to_globe(self.ellipsoid, ccrs.Globe())

        return ccrs.Geostationary(
            central_longitude=self.longitude_of_projection_origin,
            satellite_height=self.perspective_point_height,
            false_easting=self.false_easting,
            false_northing=self.false_northing,
            globe=globe,
            sweep_axis=self.sweep_angle_axis,
        )

    def as_cartopy_projection(self):
        return self.as_cartopy_crs()


class Stereographic(CoordSystem):
    """
    A stereographic map projection.

    """

    grid_mapping_name = "stereographic"

    def __init__(
        self,
        central_lat,
        central_lon,
        false_easting=None,
        false_northing=None,
        true_scale_lat=None,
        ellipsoid=None,
        *args,
        **kwargs
    ):
        """
        Constructs a Stereographic coord system.

        Args:

        * central_lat:
            The latitude of the pole.

        * central_lon:
            The central longitude, which aligns with the y axis.

        Kwargs:

        * false_easting:
            X offset from planar origin in metres. Defaults to 0.0 .

        * false_northing:
            Y offset from planar origin in metres. Defaults to 0.0 .

        * true_scale_lat:
            Latitude of true scale.

        * ellipsoid (:class:`GeogCS`):
            If given, defines the ellipsoid.

        """

        #: True latitude of planar origin in degrees.
        self.central_lat = float(central_lat)

        #: True longitude of planar origin in degrees.
        self.central_lon = float(central_lon)

        #: X offset from planar origin in metres.
        self.false_easting = _arg_default(false_easting, 0)

        #: Y offset from planar origin in metres.
        self.false_northing = _arg_default(false_northing, 0)

        #: Latitude of true scale.
        self.true_scale_lat = _arg_default(
            true_scale_lat, None, cast_as=_float_or_None
        )
        # N.B. the way we use this parameter, we need it to default to None,
        # and *not* to 0.0 .

        #: Ellipsoid definition (:class:`GeogCS` or None).
        self.ellipsoid = ellipsoid

    def __repr__(self):
        return (
            "Stereographic(central_lat={!r}, central_lon={!r}, "
            "false_easting={!r}, false_northing={!r}, "
            "true_scale_lat={!r}, "
            "ellipsoid={!r})".format(
                self.central_lat,
                self.central_lon,
                self.false_easting,
                self.false_northing,
                self.true_scale_lat,
                self.ellipsoid,
            )
        )

    def as_cartopy_crs(self):
        globe = self._ellipsoid_to_globe(self.ellipsoid, ccrs.Globe())

        return ccrs.Stereographic(
            self.central_lat,
            self.central_lon,
            self.false_easting,
            self.false_northing,
            self.true_scale_lat,
            globe=globe,
        )

    def as_cartopy_projection(self):
        return self.as_cartopy_crs()


class LambertConformal(CoordSystem):
    """
    A coordinate system in the Lambert Conformal conic projection.

    """

    grid_mapping_name = "lambert_conformal_conic"

    def __init__(
        self,
        central_lat=None,
        central_lon=None,
        false_easting=None,
        false_northing=None,
        secant_latitudes=None,
        ellipsoid=None,
        *args,
        **kwargs
    ):
        """
        Constructs a LambertConformal coord system.

        Kwargs:

        * central_lat:
                The latitude of "unitary scale".  Defaults to 39.0 .

        * central_lon:
                The central longitude.  Defaults to -96.0 .

        * false_easting:
                X offset from planar origin in metres.  Defaults to 0.0 .

        * false_northing:
                Y offset from planar origin in metres.  Defaults to 0.0 .

        * secant_latitudes (number or iterable of 1 or 2 numbers):
                Latitudes of secant intersection.  One or two.
                Defaults to (33.0, 45.0).

        * ellipsoid (:class:`GeogCS`):
            If given, defines the ellipsoid.

        .. note:

            Default arguments are for the familiar USA map:
            central_lon=-96.0, central_lat=39.0,
            false_easting=0.0, false_northing=0.0,
            secant_latitudes=(33, 45)

        """

        #: True latitude of planar origin in degrees.
        self.central_lat = _arg_default(central_lat, 39.0)

        #: True longitude of planar origin in degrees.
        self.central_lon = _arg_default(central_lon, -96.0)

        #: X offset from planar origin in metres.
        self.false_easting = _arg_default(false_easting, 0)

        #: Y offset from planar origin in metres.
        self.false_northing = _arg_default(false_northing, 0)

        #: The standard parallels of the cone (tuple of 1 or 2 floats).
        self.secant_latitudes = _arg_default(
            secant_latitudes, (33, 45), cast_as=_1or2_parallels
        )

        #: Ellipsoid definition (:class:`GeogCS` or None).
        self.ellipsoid = ellipsoid

    def __repr__(self):
        return (
            "LambertConformal(central_lat={!r}, central_lon={!r}, "
            "false_easting={!r}, false_northing={!r}, "
            "secant_latitudes={!r}, ellipsoid={!r})".format(
                self.central_lat,
                self.central_lon,
                self.false_easting,
                self.false_northing,
                self.secant_latitudes,
                self.ellipsoid,
            )
        )

    def as_cartopy_crs(self):
        # We're either north or south polar. Set a cutoff accordingly.
        if self.secant_latitudes is not None:
            lats = self.secant_latitudes
            max_lat = lats[0]
            if len(lats) == 2:
                max_lat = lats[0] if abs(lats[0]) > abs(lats[1]) else lats[1]
            cutoff = -30 if max_lat > 0 else 30
        else:
            cutoff = None

        globe = self._ellipsoid_to_globe(self.ellipsoid, ccrs.Globe())

        return ccrs.LambertConformal(
            central_longitude=self.central_lon,
            central_latitude=self.central_lat,
            false_easting=self.false_easting,
            false_northing=self.false_northing,
            globe=globe,
            cutoff=cutoff,
            standard_parallels=self.secant_latitudes,
        )

    def as_cartopy_projection(self):
        return self.as_cartopy_crs()


class Mercator(CoordSystem):
    """
    A coordinate system in the Mercator projection.

    """

    grid_mapping_name = "mercator"

    def __init__(
        self,
        longitude_of_projection_origin=None,
        ellipsoid=None,
        standard_parallel=None,
        *args,
        **kwargs
    ):
        """
        Constructs a Mercator coord system.

        Kwargs:

        * longitude_of_projection_origin:
            True longitude of planar origin in degrees. Defaults to 0.0 .

        * ellipsoid (:class:`GeogCS`):
            If given, defines the ellipsoid.

        * standard_parallel:
            The latitude where the scale is 1. Defaults to 0.0 .

        """
        #: True longitude of planar origin in degrees.
        self.longitude_of_projection_origin = _arg_default(
            longitude_of_projection_origin, 0
        )

        #: Ellipsoid definition (:class:`GeogCS` or None).
        self.ellipsoid = ellipsoid

        #: The latitude where the scale is 1.
        self.standard_parallel = _arg_default(standard_parallel, 0)

    def __repr__(self):
        res = (
            "Mercator(longitude_of_projection_origin="
            "{self.longitude_of_projection_origin!r}, "
            "ellipsoid={self.ellipsoid!r}, "
            "standard_parallel={self.standard_parallel!r})"
        )
        return res.format(self=self)

    def as_cartopy_crs(self):
        globe = self._ellipsoid_to_globe(self.ellipsoid, ccrs.Globe())

        return ccrs.Mercator(
            central_longitude=self.longitude_of_projection_origin,
            globe=globe,
            latitude_true_scale=self.standard_parallel,
        )

    def as_cartopy_projection(self):
        return self.as_cartopy_crs()


class LambertAzimuthalEqualArea(CoordSystem):
    """
    A coordinate system in the Lambert Azimuthal Equal Area projection.

    """

    grid_mapping_name = "lambert_azimuthal_equal_area"

    def __init__(
        self,
        latitude_of_projection_origin=None,
        longitude_of_projection_origin=None,
        false_easting=None,
        false_northing=None,
        ellipsoid=None,
        *args,
        **kwargs
    ):
        """
        Constructs a Lambert Azimuthal Equal Area coord system.

        Kwargs:

        * latitude_of_projection_origin:
            True latitude of planar origin in degrees. Defaults to 0.0 .

        * longitude_of_projection_origin:
            True longitude of planar origin in degrees. Defaults to 0.0 .

        * false_easting:
                X offset from planar origin in metres. Defaults to 0.0 .

        * false_northing:
                Y offset from planar origin in metres. Defaults to 0.0 .

        * ellipsoid (:class:`GeogCS`):
            If given, defines the ellipsoid.

        """
        #: True latitude of planar origin in degrees.
        self.latitude_of_projection_origin = _arg_default(
            latitude_of_projection_origin, 0
        )

        #: True longitude of planar origin in degrees.
        self.longitude_of_projection_origin = _arg_default(
            longitude_of_projection_origin, 0
        )

        #: X offset from planar origin in metres.
        self.false_easting = _arg_default(false_easting, 0)

        #: Y offset from planar origin in metres.
        self.false_northing = _arg_default(false_northing, 0)

        #: Ellipsoid definition (:class:`GeogCS` or None).
        self.ellipsoid = ellipsoid

    def __repr__(self):
        return (
            "LambertAzimuthalEqualArea(latitude_of_projection_origin={!r},"
            " longitude_of_projection_origin={!r}, false_easting={!r},"
            " false_northing={!r}, ellipsoid={!r})"
        ).format(
            self.latitude_of_projection_origin,
            self.longitude_of_projection_origin,
            self.false_easting,
            self.false_northing,
            self.ellipsoid,
        )

    def as_cartopy_crs(self):
        globe = self._ellipsoid_to_globe(self.ellipsoid, ccrs.Globe())

        return ccrs.LambertAzimuthalEqualArea(
            central_longitude=self.longitude_of_projection_origin,
            central_latitude=self.latitude_of_projection_origin,
            false_easting=self.false_easting,
            false_northing=self.false_northing,
            globe=globe,
        )

    def as_cartopy_projection(self):
        return self.as_cartopy_crs()


class AlbersEqualArea(CoordSystem):
    """
    A coordinate system in the Albers Conical Equal Area projection.

    """

    grid_mapping_name = "albers_conical_equal_area"

    def __init__(
        self,
        latitude_of_projection_origin=None,
        longitude_of_central_meridian=None,
        false_easting=None,
        false_northing=None,
        standard_parallels=None,
        ellipsoid=None,
        *args,
        **kwargs
    ):
        """
        Constructs a Albers Conical Equal Area coord system.

        Kwargs:

        * latitude_of_projection_origin:
            True latitude of planar origin in degrees. Defaults to 0.0 .

        * longitude_of_central_meridian:
            True longitude of planar central meridian in degrees.
            Defaults to 0.0 .

        * false_easting:
            X offset from planar origin in metres. Defaults to 0.0 .

        * false_northing:
            Y offset from planar origin in metres. Defaults to 0.0 .

        * standard_parallels (number or iterable of 1 or 2 numbers):
            The one or two latitudes of correct scale.
            Defaults to (20.0, 50.0).

        * ellipsoid (:class:`GeogCS`):
            If given, defines the ellipsoid.

        """
        #: True latitude of planar origin in degrees.
        self.latitude_of_projection_origin = _arg_default(
            latitude_of_projection_origin, 0
        )

        #: True longitude of planar central meridian in degrees.
        self.longitude_of_central_meridian = _arg_default(
            longitude_of_central_meridian, 0
        )

        #: X offset from planar origin in metres.
        self.false_easting = _arg_default(false_easting, 0)

        #: Y offset from planar origin in metres.
        self.false_northing = _arg_default(false_northing, 0)

        #: The one or two latitudes of correct scale (tuple of 1 or 2 floats).
        self.standard_parallels = _arg_default(
            standard_parallels, (20, 50), cast_as=_1or2_parallels
        )

        #: Ellipsoid definition (:class:`GeogCS` or None).
        self.ellipsoid = ellipsoid

    def __repr__(self):
        return (
            "AlbersEqualArea(latitude_of_projection_origin={!r},"
            " longitude_of_central_meridian={!r}, false_easting={!r},"
            " false_northing={!r}, standard_parallels={!r},"
            " ellipsoid={!r})"
        ).format(
            self.latitude_of_projection_origin,
            self.longitude_of_central_meridian,
            self.false_easting,
            self.false_northing,
            self.standard_parallels,
            self.ellipsoid,
        )

    def as_cartopy_crs(self):
        globe = self._ellipsoid_to_globe(self.ellipsoid, ccrs.Globe())

        return ccrs.AlbersEqualArea(
            central_longitude=self.longitude_of_central_meridian,
            central_latitude=self.latitude_of_projection_origin,
            false_easting=self.false_easting,
            false_northing=self.false_northing,
            standard_parallels=self.standard_parallels,
            globe=globe,
        )

    def as_cartopy_projection(self):
        return self.as_cartopy_crs()


class CurvilinearGrid(CoordSystem):
    grid_mapping_name = "curvilinear_grid"

    def __init__(self, *args, **kwargs):
        # TODO: more logic
        pass

    def as_cartopy_crs(self):
        raise NotImplementedError()

    def as_cartopy_projection(self):
        raise NotImplementedError()


def parse_crs(
    da: xr.DataArray,
) -> Optional["CoordSystem"]:
    """
    Creates an instance of CoordSystem subclass, one from those defined in this file where grid_mapping_name suits the one requested.

    Parameters
    ----------
    da : xarray.DataArray
        Xarray structure with, at least attribute `grid_mapping_name` and other attributes if required

    Returns:
    ----------
    coord_sys : CoordSystem
        Proper subclass of CoordSystem class

    Examples:
    ----------
    >>> ds = get_coord_system(
    ...         grid_mapping_name="rotated_latitude_longitude",
    ...         grid_north_pole_latitude=32.5,
    ...         grid_north_pole_longitude=170
    ...     )
    """
    grid_mapping_name = trim_prefix(da.attrs.get("grid_mapping_name"))
    if grid_mapping_name == "latitude_longitude":
        return RegularLatLon(**da.attrs)
    for _, cls in inspect.getmembers(
        importlib.import_module(__name__), inspect.isclass
    ):
        if not hasattr(cls, "grid_mapping_name"):
            continue
        if cls.grid_mapping_name == grid_mapping_name:
            return cls(**da.attrs)



from __future__ import annotations

import functools as ft
from numbers import Number
import warnings
from collections.abc import Iterable
from enum import Enum
from itertools import chain
from statistics import mode
from typing import Any, Hashable, List, Mapping, Optional, Tuple, Union

import numpy as np
import dask.array as da
import pandas as pd
import xarray as xr

from ..utils import util_methods
from ..utils.decorators import geokube_logging
from ..utils.hcube_logger import HCubeLogger
from .axis import Axis, AxisType
from .coord_system import (
    CoordSystem,
    CurvilinearGrid,
    GeogCS,
    RegularLatLon,
    parse_crs,
    add_prefix,
    trim_prefix
)
from .coordinate import Coordinate, CoordinateType
from .domainmixin import DomainMixin
from .enums import LatitudeConvention, LongitudeConvention
from .variable import Variable

_COORDS_TUPLE_CONTENT = [
    "dims",
    "data",
    "bounds",
    "units",
    "properties",
    "encoding",
]


# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring


class DomainType(Enum):
    GRIDDED = "gridded"
    POINTS = "points"
    TIMESERIES = "timeseries"


class Domain(DomainMixin):
    __slots__ = (
        "_coords",
        "_crs",
        "_type",
        "_axis_to_name",
    )

    _LOG = HCubeLogger(name="Domain")

    def __init__(
        self,
        coords: Union[
            Mapping[Hashable, Tuple[np.ndarray, ...]],
            Iterable[Coordinate],
            Domain,
        ],
        crs: CoordSystem,
        domaintype: Optional[DomainType] = None,
    ) -> None:
        if isinstance(coords, dict):
            self._coords = {}
            for name, coord in coords.items():
                self._coords[name] = Domain._as_coordinate(coord, name)
        if isinstance(coords, (list, set)):
            # TODO: check if it is a coordinate or just data!
            self._coords = {c.name: c for c in coords}
        if isinstance(coords, Domain):
            self._coords = coords._coords
            self._crs = coords._crs
            self._type = coords._type
            self._axis_to_name = coords._axis_to_name

        self._crs = crs
        self._type = domaintype
        self._axis_to_name = {
            c.axis_type: c.name for c in self._coords.values()
        }

    @classmethod
    def _as_coordinate(cls, coord, name) -> Coordinate:
        if isinstance(coord, Coordinate):
            return coord
        elif isinstance(coord, tuple):
            # tupl -> (data, dims, axis)
            l = dict(enumerate(coord))
            return Coordinate(
                data=coord[0], dims=l.get(1, name), axis=l.get(2, name)
            )
        else:
            return Coordinate(data=coord, axis=name)
    
    @property
    def grid_mapping_name(self) -> str:
        return add_prefix(self.crs.grid_mapping_name)

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, value):
        self._type = value

    @property
    def coords(self):
        return self._coords

    @property
    def crs(self) -> CoordSystem:  # horizontal coordinate reference system
        return self._crs

    @crs.setter
    def crs(self, crs):
        self._crs = crs

    @property
    def aux_coords(self) -> List[str]:
        return [c.name for c in self._coords.values() if not c.is_dim]

    def _infer_resolution(self):
        grid_x = np.abs(mode(np.diff(self.longitude)))
        grid_y = np.abs(mode(np.diff(self.latitude)))
        return (grid_x, grid_y)

    def __repr__(self) -> str:
        return self.to_xarray(encoding=False).__repr__()

    #        return formatting.array_repr(self.to_xarray())

    def _repr_html_(self):
        return self.to_xarray(encoding=False)._repr_html_()
        # if OPTIONS["display_style"] == "text":
        #     return f"<pre>{escape(repr(self.to_xarray()))}</pre>"
        # return formatting_html.array_repr(self)

    def __eq__(self, other):
        if self.crs != other.crs:
            return False
        coord_keys_eq = set(self._coords.keys()) == set(other._coords.keys())
        if not coord_keys_eq:
            return False
        for ck in self._coords.keys():
            if self._coords[ck].axis_type is AxisType.TIME:
                if not np.all(
                    self._coords[ck].values == other._coords[ck].values
                ):
                    return False
            else:
                if not np.allclose(
                    self._coords[ck].values, other._coords[ck].values
                ):
                    return False
        return True

    def __ne__(self, other):
        return not (self == other)

    def __len__(self) -> int:
        return len(self._coords)

    def __setitem__(self, key: str, value: Union[Coordinate, Variable]):
        self._coords[key] = value

    def __contains__(self, key: Union[str, Axis, AxisType]) -> bool:
        if isinstance(key, Axis):
            return key.type in self._axis_to_name
        return (key in self._coords) or (
            AxisType.parse(key) in self._axis_to_name
        )

    def __next__(self):
        for k, v in self._coords.items():
            yield k, v
        raise StopIteration

    @property
    def nbytes(self) -> int:
        return sum(coord.nbytes for coord in self._coords.values())

    def map_indexers(self, indexers: Mapping[str, Any]) -> Mapping[Axis, Any]:
        return {Axis(n): v for n, v in indexers.items()}

    @geokube_logging
    def _process_time_combo(self, indexer: Mapping[Hashable, Any]):
        if "time" in indexer:
            indexer = indexer["time"]

        def _reduce_boolean_selection(_ds, key: str, _time_indexer):
            if key in _time_indexer.keys():
                dt = getattr(_ds, key).values
                XX = ft.reduce(
                    lambda x, y: x | (dt == y),
                    [False]
                    + list(np.array(_time_indexer[key], dtype=int, ndmin=1)),
                )
                return XX
            return True

        if (time_coord := self[AxisType.TIME]) is None:
            raise KeyError(f"Time axis was not found for that dataset!")
        time_coord_dset = time_coord.to_xarray(encoding=False)
        time_coord_dt = time_coord_dset[time_coord.name].dt

        year_mask = _reduce_boolean_selection(time_coord_dt, "year", indexer)
        month_mask = _reduce_boolean_selection(time_coord_dt, "month", indexer)
        day_mask = _reduce_boolean_selection(time_coord_dt, "day", indexer)
        hour_mask = _reduce_boolean_selection(time_coord_dt, "hour", indexer)
        inds = np.where(year_mask & month_mask & day_mask & hour_mask)[0]
        inds = util_methods.list_to_slice_or_array(inds)
        return {time_coord.name: inds}

    @geokube_logging
    def compute_bounds(
        self, coordinate: str | None = None, force: bool = False
    ) -> None:
        # Defining behavior if `coordinate` is different from `'latitude'` or
        # `'longitude'`
        if coordinate is None:
            self.compute_bounds(coordinate="latitude", force=False)
            self.compute_bounds(coordinate="longitude", force=False)
            return
        elif coordinate not in {"latitude", "longitude"}:
            raise NotImplementedError(
                "'coordinate' must be either 'latitude' or 'longitude', other "
                "values are not currently supported"
            )

        # Extracting the coordinate object
        coord = self[coordinate]

        # Handling the case when bounds already exist, according to `force`
        if coord.bounds is not None:
            msg = f"{coord.name} bounds already exist"
            if not force:
                warnings.warn(f"{msg} and are not going be modified")
                self._LOG.warn(f"{msg} and are not going be modified")
                return
            warnings.warn(f"{msg} and are going to be recalculated")
            self._LOG.warn(f"{msg} and are going to be recalculated")

        # Cases for dependent or scalar coordinates are not handled
        if coord.type is not CoordinateType.INDEPENDENT:
            raise NotImplementedError(
                "'coordinate' must be independent to calculate its bounds, "
                "dependent coordinates are not currently supported"
            )

        # Handling the case when `crs` is `None` or not instance of `GeogCS`
        crs = self._crs
        if crs is None:
            # TODO: Reconsider if this should be `ValueError` or some other
            # type of exception, see:
            # https://docs.python.org/3/library/exceptions.html#ValueError
            raise ValueError(
                "'crs' is None and cell bounds cannot be calculated"
            )
        if not isinstance(crs, GeogCS):
            raise NotImplementedError(
                f"'{crs.__class__.__name__}' is currently not supported for "
                "calculating cell corners"
            )

        # Calculating bounds
        val = coord.values
        val_b = np.empty(shape=val.size + 1, dtype=np.float64)
        val_b[1:-1] = 0.5 * (val[:-1] + val[1:])
        half_step = 0.5 * (val.ptp() / (val.size - 1))
        # The case `val[0] > val[-1]` represents reversed order of values:
        i, j = (0, -1) if val[0] <= val[-1] else (-1, 0)
        val_b[i] = val[i] - half_step
        val_b[j] = val[j] + half_step
        # Making sure that longitude and latitude values are not outside their
        # ranges
        if coord.axis_type is AxisType.LONGITUDE:
            if coord.convention is LongitudeConvention.POSITIVE_WEST:
                range_b = (0.0, 360.0)
            else:
                range_b = (-180.0, 180.0)
        else:  # Case when coord.axis_type is AxisType.LATITUDE
            range_b = (-90.0, 90.0)
        val_b[i] = val_b[i].clip(*range_b)
        val_b[j] = val_b[j].clip(*range_b)

        # Setting `coordinate.bounds`
        coord.bounds = Domain.convert_bounds_1d_to_2d(val_b)

    @geokube_logging
    def _calculate_missing_lat_and_lon(self):
        # NOTE: This approach is elegant and accurate, but it issues warnings
        # because of the `@geokube_logging` decorator.
        # missing_lat_or_lon = False
        # try:
        #     self.latitude, self.longitude
        # except KeyError:
        #     missing_lat_or_lon = True
        missing_lat_or_lon = not (
            {AxisType.LATITUDE, AxisType.LONGITUDE}
            <= self._axis_to_name.keys()
        )

        # TODO: Consider moving these checks to `Field.to_xarray`.
        if (
            missing_lat_or_lon
            and (self._type is DomainType.GRIDDED or self._type is None)
            and self._crs is not None
            and not isinstance(self._crs, GeogCS)
            and self.x.type is CoordinateType.INDEPENDENT
            and self.y.type is CoordinateType.INDEPENDENT
        ):
            domain_x, domain_y = self.x, self.y
            dims = (domain_y.dims[0], domain_x.dims[0])
            x, y = np.meshgrid(domain_x.to_numpy(), domain_y.to_numpy())
            pts = (
                GeogCS(6371229)
                .as_cartopy_crs()
                .transform_points(src_crs=self._crs.as_cartopy_crs(), x=x, y=y)
            )
            lon_vals, lat_vals = pts[..., 0], pts[..., 1]
            lat_axis = Axis(
                name="latitude", axistype=AxisType.LATITUDE, is_dim=False
            )
            lon_axis = Axis(
                name="longitude", axistype=AxisType.LONGITUDE, is_dim=False
            )
            lat_coord = Coordinate(
                data=lat_vals, axis=lat_axis, dims=dims, units="degree_north"
            )
            lon_coord = Coordinate(
                data=lon_vals, axis=lon_axis, dims=dims, units="degree_east"
            )
            new_coords = {"latitude": lat_coord, "longitude": lon_coord}
            self._coords.update(new_coords)
            new_names = {
                AxisType.LATITUDE: "latitude",
                AxisType.LONGITUDE: "longitude",
            }
            self._axis_to_name.update(new_names)

    @staticmethod
    def convert_bounds_1d_to_2d(values):
        assert values.ndim == 1
        return np.vstack((values[:-1], values[1:])).T

    @staticmethod
    def convert_bounds_2d_to_1d(values):
        assert values.ndim == 2
        return np.concatenate((values[:, 0], values[[-1], 1]))

    @classmethod
    def guess_crs(
        cls,
        da: Union[xr.Dataset, xr.DataArray, Mapping[str, Coordinate]],
    ):
        # TODO: implement more logic
        if isinstance(da, (xr.Dataset, xr.DataArray)):
            if "nav_lat" in da.coords or "nav_lon" in da.coords:
                return CurvilinearGrid()
        if isinstance(da, dict):
            if "nav_lat" in da or "nav_lon" in da:
                return CurvilinearGrid()
        return GeogCS(6371229)

    @classmethod
    @geokube_logging
    def merge(cls, domains: List[Domain]):
        # TODO: check if the domains are defined on the same crs
        coords = {}
        domaintype = None
        for domain in domains:
            coords.update(**domain.coords)
            domaintype = domain.type
        return Domain(coords=coords, crs=domains[0].crs, domaintype=domaintype)

    @classmethod
    @geokube_logging
    def from_xarray(
        cls,
        ds: xr.Dataset,
        ncvar: str,
        id_pattern: str = None,
        copy: bool = False,
        mapping: Optional[Mapping[str, str]] = None,
    ) -> "Domain":
        da = ds[ncvar]
        coords = set()
        for dim_name in da.dims:
            if dim_name in da.coords:
                coords.add(
                    Coordinate.from_xarray(
                        ds=ds,
                        ncvar=dim_name,
                        id_pattern=id_pattern,
                        mapping=mapping,
                    )
                )
        xr_coords = ds[ncvar].attrs.get(
            "coordinates", ds[ncvar].encoding.get("coordinates", None)
        )
        if xr_coords is not None:
            for coord_name in xr_coords.split(" "):
                if coord_name not in ds:
                    warnings.warn(
                        f"Coordinate {coord_name} does not exist in the"
                        " dataset!"
                    )
                    continue
                coord = Coordinate.from_xarray(
                    ds=ds,
                    ncvar=coord_name,
                    id_pattern=id_pattern,
                    mapping=mapping,
                )
                if coord in coords:
                    warnings.warn(
                        f"Coordinate {coord_name} was already defined as"
                        " dimension!"
                    )
                    continue
                coords.add(coord)
        if "grid_mapping" in da.encoding:
            crs = parse_crs(da[da.encoding.get("grid_mapping")])
        elif "grid_mapping" in da.attrs:
            crs = parse_crs(da[da.attrs.get("grid_mapping")])
        else:
            crs = Domain.guess_crs(da)

        # NOTE: a workaround for keeping domaintype
        # ds attributes are modified!
        # Issue: https://github.com/geokube/geokube/issues/147
        if (
            domain_type := ds[ncvar].attrs.pop("__geo_domtype", None)
        ) is not None:
            return Domain(
                coords=coords, crs=crs, domaintype=DomainType(domain_type)
            )
        return Domain(coords=coords, crs=crs)

    @geokube_logging
    def to_xarray(
        self, encoding=True
    ) -> xr.core.coordinates.DatasetCoordinates:

        grid = {}
        #grid = xr.Dataset(coords=self._coords.).coords
        for coord in self._coords.values():
            var_name = coord.ncvar if encoding else coord.name
            grid[var_name] = coord.to_xarray(encoding=encoding)[var_name]
        #grid = xr.Dataset(coords=grid).coords
        if self.crs is not None:
            crs_name = self.grid_mapping_name
            not_none_attrs = self.crs.as_crs_attributes()
            # NOTE: we keep cf-compliant value of `grid_mapping_name` attribute
            not_none_attrs["grid_mapping_name"] = self.crs.grid_mapping_name
            grid.update(
                {
                    crs_name: xr.DataArray(
                        1, name=crs_name, attrs=not_none_attrs
                    )
                }
            )
        return xr.Dataset(coords=grid).coords

    def to_dict(self, unique_values=False):
        return {
            "crs": self._crs.to_dict(),
            "coordinates": {
                name: coord.to_dict(unique_values)
                for name, coord in self._coords.items()
            },
        }

    @classmethod
    def _make_domain_from_coords_dict_dims_and_crs(
        cls, coords, dims, crs=None
    ):
        """Return a domain based on coords dict, dims, and coordinate reference system.

        coords can be in the form {"latitude": lat_value} or in the form where the value
        is a tuple. That tuple might contain following elements:
        (dims: tuple[str], data, unit: optional, bounds: optional, properties: optional, encoding: optional)

        """
        if not isinstance(coords, dict):
            raise TypeError(
                f"Expected type of `coords` is `dict`, but `{type(coords)}`"
                " provided!"
            )
        res_coords = []
        for k, v in coords.items():
            if isinstance(v, pd.core.indexes.datetimes.DatetimeIndex):
                v = np.array(v)
            if isinstance(v, (Number, np.ndarray, da.Array)):
                # If coord provided not as tuple, be default it is deemed as `dimension`
                if isinstance(k, AxisType):
                    axis = Axis(name=k.axis_type_name, axistype=k, is_dim=True)
                elif isinstance(k, (Axis, str)):
                    axis = Axis(k, is_dim=True)
                res_coords.append(Coordinate(data=v, axis=axis, dims=(axis,)))
            elif isinstance(v, tuple):
                dims = data = bounds = units = props = encoding = None
                if len(v) >= 1:
                    dims = v[0]
                if len(v) >= 2:
                    data = v[1]
                if len(v) >= 3:
                    bounds = {k: v[2]}
                if len(v) >= 4:
                    units = v[3]
                if len(v) >= 5:
                    props = v[4]
                if len(v) >= 6:
                    encoding = v[5]
                res_coords.append(
                    Coordinate(
                        data=data,
                        axis=Axis(k),
                        dims=dims,
                        bounds=bounds,
                        units=units,
                        properties=props,
                        encoding=encoding,
                    )
                )
            else:
                raise TypeError(
                    "Expected types of coord values are following: [Number,"
                    " numpy.ndarray, dask.array.Array, tuple], but proided"
                    f" type was `{type(v)}`"
                )

        if crs is None:
            crs = Domain.guess_crs(coords)

        return Domain(coords=res_coords, crs=crs)


class GeodeticPoints(Domain):
    def __init__(self, latitude, longitude, vertical=None):
        latitude = np.array(latitude, dtype=np.float64, ndmin=1)
        longitude = np.array(longitude, dtype=np.float64, ndmin=1)
        if vertical != None:
            vertical = np.array(vertical, dtype=np.float64, ndmin=1)
            super().__init__(
                coords={
                    "latitude": (latitude, "points", "latitude"),
                    "longitude": (longitude, "points", "longitude"),
                    "vertical": (vertical, "points", "vertical"),
                },
                crs=GeogCS(6371229),
                domaintype=DomainType.POINTS,
            )
        else:
            super().__init__(
                coords={
                    "latitude": (latitude, "points", "latitude"),
                    "longitude": (longitude, "points", "longitude"),
                },
                crs=GeogCS(6371229),
                domaintype=DomainType.POINTS,
            )


class GeodeticGrid(Domain):
    def __init__(self, latitude, longitude, vertical=None):
        latitude = np.array(latitude, dtype=np.float64, ndmin=1)
        longitude = np.array(longitude, dtype=np.float64, ndmin=1)
        if vertical != None:
            vertical = np.array(vertical, dtype=np.float64, ndmin=1)
            super().__init__(
                coords={
                    "latitude": latitude,
                    "longitude": longitude,
                    "vertical": vertical,
                },
                crs=GeogCS(6371229),
            )
        else:
            # TODO: TO BE FIXED
            super().__init__(
                coords={
                    "latitude": (latitude, "latitude"),
                    "longitude": (longitude, "longitude"),
                },
                crs=GeogCS(6371229),
            )



from .variable import Variable


class Bounds(Variable):
    __slots__ = ("_name",)

    @property
    def name(self):
        return self._name


class Bounds1D(Bounds):
    pass


class BoundsND(Bounds):
    pass



from collections.abc import Iterable
from enum import Enum
from numbers import Number
from typing import Any, Hashable, Mapping, Optional, Union

import dask.array as da
import numpy as np
import xarray as xr


class Types(Enum):
    INT = "int"
    STR = "str"
    BOOL = "bool"
    EMPTY = "empty"
    UNKNOWN = "unknown"


BoolTypes = (bool, np.bool_)
IntTypes = (int, np.int8, np.int16, np.int32, np.int64)
CollectionTypes = Iterable


def infer_type(key: Any) -> Types:
    if isinstance(key, slice) and (
        isinstance(key.start, IntTypes) or isinstance(key.stop, IntTypes)
    ):
        return Types.INT
    if isinstance(key, CollectionTypes) and not isinstance(key, str):
        try:
            f = next(iter(key))
            if isinstance(f, BoolTypes):
                return Types.BOOL
            if isinstance(f, IntTypes):
                return Types.INT
            if isinstance(f, str):
                return Types.STR
            return Types.UNKNOWN
        except StopIteration:
            return Types.EMPTY
    if isinstance(key, BoolTypes):
        return Types.BOOL
    if isinstance(key, IntTypes):
        return Types.INT
    if isinstance(key, str):
        return Types.STR
    return Types.UNKNOWN



import cf_units as cf


class Unit:
    UNKNOWN_CF_UNIT = cf.Unit(None)

    __slots__ = (
        "_unit",
        "_backup_name",
    )

    def __init__(self, unit, calendar=None):
        try:
            self._unit = cf.Unit(unit=unit, calendar=calendar)
            self._backup_name = None
        except ValueError:
            self._unit = cf.Unit(unit=None, calendar=calendar)
            self._backup_name = unit

    @property
    def is_unknown(self):
        return (
            self._backup_name is not None or self._unit == Unit.UNKNOWN_CF_UNIT
        )

    def __repr__(self):
        return f"<Unit({str(self._unit)}, backeup_name={self._backup_name}>"

    def __str__(self):
        if self._backup_name is None:
            return str(self._unit)
        return self._backup_name

    def __getattr__(self, name):
        if not hasattr(self._unit, name):
            raise AttributeError(f"Attribute `{name}` is not available.")
        return getattr(self._unit, name)

    def __getstate__(self):
        return dict(
            **self._unit.__getstate__(), **{"backup_name": self._backup_name}
        )

    def __setstate__(self, state):
        self.__init__(state["unit_text"], calendar=state["calendar"])

    def __eq__(self, other):
        if not isinstance(other, Unit):
            return False
        return (
            self._unit == other._unit
            and self._backup_name == other._backup_name
        )

    def __ne__(self, other):
        return not (self == other)



from numbers import Number

import dask.array as da
import numpy as np
import dask.array as da
import pytest

from geokube.core.axis import Axis, AxisType
from geokube.core.coordinate import Coordinate, CoordinateType
from geokube.core.unit import Unit
from geokube.core.variable import Variable
from tests import compare_dicts
from tests.fixtures import *


def test_process_bounds_fails():
    with pytest.raises(
        TypeError,
        match=(
            r"Expected argument is one of the following types `dict`,"
            r" `numpy.ndarray`, or `geokube.Variable`, but provided*"
        ),
    ):
        Coordinate._process_bounds(
            [1, 2, 3, 4],
            name="name",
            units="units",
            axis=Axis("time"),
            variable_shape=(100, 1),
        )

    with pytest.raises(
        TypeError,
        match=(
            r"Expected argument is one of the following types `dict`,"
            r" `numpy.ndarray`, or `geokube.Variable`, but provided*"
        ),
    ):
        Coordinate._process_bounds(
            "bounds",
            name="name",
            units="units",
            axis=Axis("time"),
            variable_shape=(100, 1),
        )

    with pytest.raises(
        TypeError,
        match=(
            r"Expected argument is one of the following types `dict`,"
            r" `numpy.ndarray`, or `geokube.Variable`, but provided*"
        ),
    ):
        Coordinate._process_bounds(
            xr.DataArray([1, 2, 3, 4]),
            name="name",
            units="units",
            axis=Axis("time"),
            variable_shape=(100, 1),
        )

    with pytest.raises(ValueError, match=r"Bounds should *"):
        _ = Coordinate._process_bounds(
            Variable(data=np.random.rand(100, 5), dims=["time", "bounds"]),
            name="name",
            units="units",
            axis=Axis("time"),
            variable_shape=(100, 1),
        )

    with pytest.raises(ValueError, match=r"Bounds should *"):
        _ = Coordinate._process_bounds(
            Variable(
                data=np.random.rand(100, 5), dims=["time", "bounds"], units="m"
            ),
            name="name",
            units="units",
            axis=Axis("time"),
            variable_shape=(100, 1),
        )

    with pytest.raises(ValueError, match=r"Bounds should *"):
        _ = Coordinate._process_bounds(
            bounds=da.random.random((400, 2)),
            name="name",
            units="units",
            axis=Axis("time"),
            variable_shape=(100, 1),
        )


def test_process_bounds_proper_attrs_setting():
    D = np.random.rand(100, 2)
    r = Coordinate._process_bounds(
        Variable(data=D, dims=["time", "bounds"]),
        name="name",
        units="units",
        axis=Axis("time"),
        variable_shape=(100, 1),
    )
    assert isinstance(r, dict)
    assert "name_bounds" in r
    assert np.all(r["name_bounds"] == D)

    r = Coordinate._process_bounds(
        bounds=D,
        name="name",
        units="units",
        axis=Axis("time"),
        variable_shape=(100, 1),
    )
    assert isinstance(r, dict)
    assert "name_bounds" in r
    assert isinstance(r["name_bounds"], Variable)
    assert r["name_bounds"].units == Unit("units")
    assert np.all(r["name_bounds"] == D)

    D = da.random.random((400, 2))
    r = Coordinate._process_bounds(
        bounds=D,
        name="name2",
        units="units",
        axis=Axis("time"),
        variable_shape=(400, 1),
    )
    assert isinstance(r, dict)
    assert "name2_bounds" in r
    assert isinstance(r["name2_bounds"], Variable)
    assert r["name2_bounds"].units == Unit("units")
    assert np.all(r["name2_bounds"] == D)


def test_process_bounds_using_dict():
    d = {
        "q": np.ones((100, 2)),
        "w_bounds": Variable(
            data=np.full((100, 2), fill_value=10), dims=("lat", "bounds")
        ),
    }

    # with pytest.raises(ValueError, match=r"Bounds should*"):
    #     _ = Coordinate._process_bounds(
    #         d, name="name2", units="m", axis=Axis("lat"), variable_shape=(400, 1)
    #     )

    r = Coordinate._process_bounds(
        d, name="name2", units="m", axis=Axis("lon"), variable_shape=(100, 1)
    )
    assert "q" in r
    assert "w_bounds" in r
    assert r["q"].units == Unit("m")
    assert set(r["q"].dim_names) == {"lon", "bounds"}
    assert r["w_bounds"].units == Unit(
        None
    )  # no unit provided in `w` Variable definition
    assert set(r["w_bounds"].dim_names) == {
        "lat",
        "bounds",
    }  # `lat` defined as `w` Variable dim


def test_init_fails():
    with pytest.raises(ValueError, match=r"`data` cannot be `None`"):
        _ = Coordinate(data=None, axis="lat")

    with pytest.raises(
        TypeError,
        match=(
            r"Expected argument is one of the following types `geokube.Axis`"
            r" or `str`, but provided *"
        ),
    ):
        _ = Coordinate(data=np.ones(100), axis=["lat"])

    with pytest.raises(
        TypeError,
        match=(
            r"Expected argument is one of the following types `geokube.Axis`"
            r" or `str`, but provided *"
        ),
    ):
        _ = Coordinate(data=np.ones(100), axis=15670)

    with pytest.raises(
        ValueError,
        match=(
            r"If coordinate is not a dimension, you need to supply `dims`"
            r" argument!"
        ),
    ):
        _ = Coordinate(data=np.ones(100), axis=Axis("lat", is_dim=False))


def test_init_from_dask():
    D = da.random.random((100,))
    c = Coordinate(data=D, axis=Axis("latitude", is_dim=True), dims="latitude")
    assert c.dim_names == ("latitude",)
    assert c.dim_ncvars == ("latitude",)
    assert c.type is CoordinateType.INDEPENDENT
    assert c.axis_type is AxisType.LATITUDE
    assert c.is_independent


@pytest.mark.skip(
    "Invalidate as in the current version, if `dims` is None, it is created"
    " based on provided `axis`"
)
def test_init_from_dask_fail():
    D = da.random.random((100,))

    with pytest.raises(ValueError, match=r"Provided data have *"):
        _ = Coordinate(
            data=D,
            axis=Axis("lon", is_dim=True, encoding={"name": "new_lon_name"}),
            dims=None,
        )


def test_init_from_dask_proper_attrs_setting():
    D = da.random.random((100,))
    c = Coordinate(
        data=D,
        axis=Axis("lon", is_dim=True, encoding={"name": "new_lon_name"}),
        dims="lon",
    )
    assert c.name == "lon"
    assert c.ncvar == "new_lon_name"
    assert c.dim_names == ("lon",)
    assert c.dim_ncvars == ("lon",)
    assert c.type is CoordinateType.INDEPENDENT
    assert c.axis_type is AxisType.LONGITUDE
    assert c.is_independent


def test_init_from_dask_fail_on_bounds_shape():
    D = da.random.random((100,))
    with pytest.raises(ValueError, match=r"Bounds should*"):
        _ = Coordinate(
            data=D,
            axis=Axis("lon", is_dim=True, encoding={"name": "new_lon_name"}),
            bounds=np.ones((104, 2)),
        )


def test_init_from_numpy_proper_attrs_setting():
    D = np.random.random((100,))
    c = Coordinate(
        data=D,
        axis=Axis("lon", is_dim=True, encoding={"name": "new_lon_name"}),
        dims="lon",
        encoding={"name": "my_lon_name"},
    )
    assert c.name == "lon"
    # if encoding provieded for Axis and Cooridnate, they are merged. Keys in Axis encoding will be overwritten
    assert c.ncvar == "my_lon_name"
    assert c.dim_names == ("lon",)
    assert c.dim_ncvars == ("lon",)
    assert c.type is CoordinateType.INDEPENDENT
    assert c.axis_type is AxisType.LONGITUDE
    assert c.is_independent


def test_init_with_scalar_data():
    c = Coordinate(data=10, axis="longitude")
    assert c.type is CoordinateType.SCALAR
    assert c.axis_type is AxisType.LONGITUDE
    assert np.all(c.values == 10)
    assert c.is_independent  # Scalar treated as independent


def test_init_fails_on_missing_dim():
    D = da.random.random((100, 50))
    with pytest.raises(ValueError, match=r"Provided data have *"):
        _ = Coordinate(data=D, axis="longitude", dims="longitude")


def test_init_proper_multidim_coord():
    D = da.random.random((100, 50))
    c = Coordinate(data=D, axis="longitude", dims=["x", "y"])
    assert c.type is CoordinateType.DEPENDENT
    assert c.axis_type is AxisType.LONGITUDE
    assert c.dim_names == ("x", "y")
    assert c.dim_ncvars == ("x", "y")
    assert c.is_dependent


def test_from_xarray__regular_latlon(era5_netcdf):
    c = Coordinate.from_xarray(era5_netcdf, "time")
    assert c.type is CoordinateType.INDEPENDENT
    assert c.axis_type is AxisType.TIME
    assert c.dim_names == ("time",)
    assert c.units == Unit(
        era5_netcdf["time"].encoding["units"],
        era5_netcdf["time"].encoding["calendar"],
    )
    assert c.bounds is None
    assert not c.has_bounds


def test_from_xarray_rotated_pole(era5_rotated_netcdf):
    c = Coordinate.from_xarray(era5_rotated_netcdf, "soil1")
    assert c.dim_names == ("depth",)
    assert c.dim_ncvars == ("soil1",)
    assert c.has_bounds
    assert c.bounds is not None
    assert c.name == "depth"
    assert c.ncvar == "soil1"
    assert c.type is CoordinateType.INDEPENDENT
    assert c.axis_type is AxisType.VERTICAL or c.axis_type is AxisType.Z
    assert c.dim_names == ("depth",)
    assert c.dim_ncvars == ("soil1",)
    assert c.units == Unit("m")
    assert c.bounds["soil1_bnds"].units == Unit("m")


def test_from_xarray_rotated_pole_with_mapping(era5_rotated_netcdf):
    c = Coordinate.from_xarray(
        era5_rotated_netcdf, "soil1", mapping={"soil1": {"name": "new_soil"}}
    )
    assert c.has_bounds
    assert c.bounds is not None
    assert c.name == "new_soil"
    assert c.ncvar == "soil1"
    assert c.dim_names == ("new_soil",)
    assert c.dim_ncvars == ("soil1",)
    assert c.type is CoordinateType.INDEPENDENT
    assert c.axis_type is AxisType.VERTICAL or c.axis_type is AxisType.Z
    assert c.dim_ncvars == ("soil1",)
    assert c.units == Unit("m")
    assert c.bounds["soil1_bnds"].units == Unit("m")


def test_to_xarray_rotated_pole_without_encoding(era5_rotated_netcdf):
    c = Coordinate.from_xarray(era5_rotated_netcdf, "soil1")
    coord_dset = c.to_xarray(encoding=False)
    coord = coord_dset["depth"]

    assert coord.name == "depth"
    assert np.all(era5_rotated_netcdf.soil1.values == coord.depth.values)
    assert coord.attrs == era5_rotated_netcdf.soil1.attrs

    assert ("bounds" in coord.encoding) or ("bounds" in coord.attrs)
    compare_dicts(
        coord.encoding,
        era5_rotated_netcdf.soil1.encoding,
        exclude_d1=["name", "_FillValue"],
        exclude_d2="_FillValue",
    )
    assert coord.encoding["_FillValue"] is None


def test_to_xarray_rotated_pole_with_encoding(era5_rotated_netcdf):
    c = Coordinate.from_xarray(era5_rotated_netcdf, "soil1")
    coord_dset = c.to_xarray(encoding=True)
    coord = coord_dset["soil1"]

    assert coord.name == "soil1"
    assert np.all(era5_rotated_netcdf.soil1.values == coord.soil1.values)
    assert coord.attrs == era5_rotated_netcdf.soil1.attrs
    compare_dicts(
        coord.encoding,
        era5_rotated_netcdf.soil1.encoding,
        exclude_d1=["name", "_FillValue"],
        exclude_d2="_FillValue",
    )
    assert coord.encoding["_FillValue"] is None


def test_to_xarray_rotated_pole_with_encoding_2(era5_rotated_netcdf):
    c = Coordinate.from_xarray(era5_rotated_netcdf, "lat")
    assert c.type is CoordinateType.DEPENDENT
    coord_dset = c.to_xarray(encoding=False)
    coord = coord_dset["latitude"]
    assert coord.name == "latitude"
    assert "grid_latitude" in coord.dims
    assert "grid_longitude" in coord.dims
    assert np.all(era5_rotated_netcdf.lat.values == coord.latitude.values)
    assert coord.attrs == era5_rotated_netcdf.lat.attrs
    assert set(coord.encoding) - {"name", "_FillValue"} == set(
        era5_rotated_netcdf.lat.encoding.keys()
    ) - {"bounds"}


def test_to_xarray_rotated_pole_without_encoding_2(era5_rotated_netcdf):
    c = Coordinate.from_xarray(era5_rotated_netcdf, "lat")
    coord_dset = c.to_xarray(encoding=True)
    coord = coord_dset["lat"]
    assert coord.name == "lat"
    assert "rlat" in coord.dims
    assert "rlon" in coord.dims
    assert np.all(era5_rotated_netcdf.lat.values == coord.lat.values)
    assert coord.attrs == era5_rotated_netcdf.lat.attrs
    assert set(coord.encoding) - {"name", "_FillValue"} == set(
        era5_rotated_netcdf.lat.encoding.keys()
    ) - {"bounds"}


def test_toxarray_keeping_encoding_encoding_false_no_dims_passed():
    D = da.random.random((100,))
    c = Coordinate(data=D, axis=Axis("lat", is_dim=True))
    coord_dset = c.to_xarray(encoding=False)
    coord = coord_dset["lat"]
    assert "lat" in coord.coords
    assert "latitude" not in coord.coords
    assert coord.name == "lat"
    assert coord.dims == ("lat",)
    assert "standard_name" in coord.attrs
    assert coord.attrs["standard_name"] == "latitude"
    assert "units" in coord.attrs
    assert coord.attrs["units"] == "degrees_north"
    assert "name" in coord.encoding
    assert coord.encoding["name"] == "lat"

    c = Coordinate(data=D, axis=Axis("latitude", is_dim=True))
    coord_dset = c.to_xarray(encoding=False)
    coord = coord_dset["latitude"]
    assert "latitude" in coord.coords
    assert coord.name == "latitude"
    assert "lat" not in coord.coords
    assert coord.dims == ("latitude",)
    assert "standard_name" in coord.attrs
    assert coord.attrs["standard_name"] == "latitude"
    assert "units" in coord.attrs
    assert coord.attrs["units"] == "degrees_north"
    assert "name" in coord.encoding
    assert coord.encoding["name"] == "latitude"


def test_toxarray_keeping_encoding_encoding_false():
    D = da.random.random((100,))
    c = Coordinate(data=D, axis=Axis("lat", is_dim=True), dims="lat")
    coords_dset = c.to_xarray(encoding=False)
    coord = coords_dset["lat"]
    assert "lat" in coord.coords
    assert coord.name == "lat"
    assert coord.dims == ("lat",)
    assert "standard_name" in coord.attrs
    assert coord.attrs["standard_name"] == "latitude"
    assert "units" in coord.attrs
    assert coord.attrs["units"] == "degrees_north"
    assert "name" in coord.encoding
    assert coord.encoding["name"] == "lat"

    c = Coordinate(data=D, axis=Axis("latitude", is_dim=True), dims="latitude")
    coords_dset = c.to_xarray(encoding=False)
    coord = coords_dset["latitude"]
    assert coord.dims == ("latitude",)
    assert coord.name == "latitude"
    assert "standard_name" in coord.attrs
    assert coord.attrs["standard_name"] == "latitude"
    assert "units" in coord.attrs
    assert coord.attrs["units"] == "degrees_north"
    assert "name" in coord.encoding
    assert coord.encoding["name"] == "latitude"


def test_init_fails_if_is_dim_and_axis_name_differ_from_dims():
    D = da.random.random((100,))
    with pytest.raises(
        ValueError,
        match=(
            r"If the Coordinate is a dimension, it has to depend only on"
            r" itself, but provided `dims` are*"
        ),
    ):
        _ = Coordinate(data=D, axis=Axis("lat", is_dim=True), dims=("x", "y"))

    with pytest.raises(
        ValueError,
        match=(
            r"`dims` parameter for dimension coordinate should have the same"
            r" name as axis name*"
        ),
    ):
        _ = Coordinate(data=D, axis=Axis("lat", is_dim=True), dims="latitude")

    with pytest.raises(
        ValueError,
        match=(
            r"`dims` parameter for dimension coordinate should have the same"
            r" name as axis name*"
        ),
    ):
        _ = Coordinate(data=D, axis=Axis("latitude", is_dim=True), dims="lat")


def test_toxarray_keeping_encoding_encoding_true():
    D = da.random.random((100,))
    c = Coordinate(data=D, axis=Axis("lat", is_dim=True))
    coord_dset = c.to_xarray(encoding=True)
    coord = coord_dset["lat"]
    assert coord.name == "lat"
    assert coord.dims == ("lat",)
    assert "standard_name" in coord.attrs
    assert coord.attrs["standard_name"] == "latitude"
    assert "units" in coord.attrs
    assert coord.attrs["units"] == "degrees_north"
    assert "name" in coord.encoding
    assert coord.encoding["name"] == "lat"

    c = Coordinate(data=D, axis=Axis("latitude", is_dim=True), dims="latitude")
    coord_dset = c.to_xarray(encoding=True)
    coord = coord_dset["latitude"]
    assert coord.name == "latitude"
    assert coord.dims == ("latitude",)
    assert "standard_name" in coord.attrs
    assert coord.attrs["standard_name"] == "latitude"
    assert "units" in coord.attrs
    assert coord.attrs["units"] == "degrees_north"
    assert "name" in coord.encoding
    assert coord.encoding["name"] == "latitude"


def test_coord_data_always_numpy_array(era5_rotated_netcdf, era5_netcdf):
    for c in era5_rotated_netcdf.coords.keys():
        coord = Coordinate.from_xarray(era5_rotated_netcdf, c)
        assert isinstance(coord._data, np.ndarray)

    for c in era5_netcdf.coords.keys():
        coord = Coordinate.from_xarray(era5_netcdf, c)
        assert isinstance(coord._data, np.ndarray)

    D = da.random.random((100,))
    coord = Coordinate(
        data=D,
        axis=Axis("lon2", is_dim=True, encoding={"name": "new_lon_name"}),
        dims="lon2",
        encoding={"name": "my_lon_name"},
    )

    assert isinstance(coord._data, np.ndarray)

    D = np.random.random((100,))
    coord = Coordinate(
        data=D,
        axis=Axis("lon", is_dim=True, encoding={"name": "new_lon_name"}),
        dims="lon",
        encoding={"name": "my_lon_name"},
    )

    assert isinstance(coord._data, np.ndarray)


def test_vertical_pattern_model_level_number(nemo_ocean_16):
    nemo_ocean_16.depthv.attrs["standard_name"] = "model_level_number"
    coord = Coordinate.from_xarray(nemo_ocean_16, "depthv")
    assert coord.axis_type is AxisType.VERTICAL


def test_to_xarray_with_bounds(era5_rotated_netcdf, nemo_ocean_16):
    coord = Coordinate.from_xarray(era5_rotated_netcdf, "time")
    coord_dset = coord.to_xarray(encoding=False)
    assert "time_bnds" in coord_dset
    assert "bounds" in coord_dset["time"].encoding
    assert coord_dset["time"].encoding["bounds"] == "time_bnds"

    coord = Coordinate.from_xarray(nemo_ocean_16, "time_counter")
    coord_dset = coord.to_xarray(encoding=False)
    assert "time_counter_bounds" in coord_dset
    assert "bounds" in coord_dset["time"].encoding
    assert coord_dset["time"].encoding["bounds"] == "time_counter_bounds"

    coord_dset = coord.to_xarray(encoding=True)
    assert "time_counter_bounds" in coord_dset
    assert "bounds" in coord_dset["time_counter"].encoding
    assert (
        coord_dset["time_counter"].encoding["bounds"] == "time_counter_bounds"
    )

    coord = Coordinate.from_xarray(nemo_ocean_16, "nav_lat")
    coord_dset = coord.to_xarray(encoding=False)
    assert "bounds_lat" in coord_dset
    assert "bounds" in coord_dset["latitude"].encoding
    assert coord_dset["latitude"].encoding["bounds"] == "bounds_lat"

    coord_dset = coord.to_xarray(encoding=True)
    assert "bounds_lat" in coord_dset
    assert "bounds" in coord_dset["nav_lat"].encoding
    assert coord_dset["nav_lat"].encoding["bounds"] == "bounds_lat"

    coord = Coordinate.from_xarray(nemo_ocean_16, "nav_lon")
    coord_dset = coord.to_xarray(encoding=False)
    assert "bounds_lon" in coord_dset
    assert "bounds" in coord_dset["longitude"].encoding
    assert coord_dset["longitude"].encoding["bounds"] == "bounds_lon"

    coord_dset = coord.to_xarray(encoding=True)
    assert "bounds_lon" in coord_dset
    assert "bounds" in coord_dset["nav_lon"].encoding
    assert coord_dset["nav_lon"].encoding["bounds"] == "bounds_lon"


def test_era5_check_if_independent_when_name_encoding_set(era5_netcdf):
    era5_netcdf["latitude"].encoding["name"] = "lat"
    coord = Coordinate.from_xarray(era5_netcdf, "latitude")
    assert coord.is_independent


def test_to_dict_store_proper_keys(nemo_ocean_16):
    details = Coordinate.from_xarray(nemo_ocean_16, "nav_lon").to_dict()
    #print(details)
    assert isinstance(details, dict)
    #assert "values" in details
    assert "units" in details
    assert "axis" in details


def test_to_dict_use_standard_name(nemo_ocean_16):
    details = Coordinate.from_xarray(nemo_ocean_16, "nav_lon").to_dict()
    assert isinstance(details, dict)
    assert details["units"] == "degrees_east"
    assert details["axis"] == "LONGITUDE"


def test_to_dict_not_store_all_values(nemo_ocean_16):
    details = Coordinate.from_xarray(nemo_ocean_16, "nav_lon").to_dict()
    assert isinstance(details, dict)
    assert "values" not in details.keys()

@pytest.mark.skip(
    "Invalidate as in the current version, Coordinate does not contain data values"
)
def test_to_dict_store_unique_values(nemo_ocean_16):
    details = Coordinate.from_xarray(nemo_ocean_16, "nav_lon").to_dict()
    #assert isinstance(details["values"], list)
    assert np.all(
        np.array(details["values"]) == np.unique(nemo_ocean_16.nav_lon.values)
    )



from enum import Enum
from typing import List, Mapping, Tuple


class CFAttributes(Enum):
    # encoding for names and dims
    NETCDF_NAME = "name"
    NETCDF_DIMS = "dims"

    # Description of data:
    UNITS = "units"
    STANDARD_NAME = "standard_name"
    LONG_NAME = "long_name"
    ANCILLARY_VARIABLES = " ancillary_variables"
    MISSING_VALUE = "missing_value"
    VALID_RANGE = "valid_range"
    VALID_MIN = "valid_min"
    VALID_MAX = "valid_max"
    FLAG_VALUES = "flag_values"
    FLAG_MEANINGS = "flag_meanings"
    FLAG_MASKS = "flag_masks"
    FILL_VALUE = "_FillValue"

    # Coordinate systems
    COORDINATES = "coordinates"
    AXIS = "axis"
    BOUNDS = "bounds"
    GRID_MAPPING = "grid_mapping"
    FORMULA_TERMS = "formula_terms"
    CALENDAR = "calendar"
    POSITIVE = "positive"

    # Data packing
    ADD_OFFSET = "add_offset"
    SCALE_FACTOR = "scale_factor"
    COMPRESS = "compress"

    # Data cell properties and methods
    CELL_MEASURES = "cell_measures"
    CELL_METHODS = "cell_methods"
    CLIMATOLOGY = "climatology"

    @classmethod
    def get_names(cls) -> List[str]:
        return [a.value for a in cls]

    @classmethod
    def split_to_props_encoding(
        cls, attrs: Mapping[str, str]
    ) -> Tuple[Mapping[str, str], Mapping[str, str]]:
        properties = attrs.copy()
        cf_encoding = {
            k: properties.pop(k) for k in cls.get_names() if k in attrs
        }
        return (properties, cf_encoding)


ENCODING_PROP = (
    "source",
    "dtype",
    "original_shape",
    "chunksizes",
    "zlib",
    "shuffle",
    "complevel",
    "fletcher32",
    "contiguous",
    CFAttributes.COORDINATES.value,
    CFAttributes.CALENDAR.value,
    CFAttributes.GRID_MAPPING.value,
    CFAttributes.MISSING_VALUE.value,
    CFAttributes.FILL_VALUE.value,
    CFAttributes.SCALE_FACTOR.value,
    CFAttributes.ADD_OFFSET.value,
)


def is_time_unit(unit):
    return "since" in unit if isinstance(unit, str) else False


def in_encoding(key, unit=None):
    return is_time_unit(unit) or key in ENCODING_PROP


def split_to_xr_attrs_and_encoding(
    mapping: Mapping[str, str]
) -> Tuple[Mapping[str, str], Mapping[str, str]]:
    attrs, encoding = {}, {}
    if mapping is not None:
        for k, v in mapping.items():
            if in_encoding(k, v):
                encoding[k] = v
            else:
                attrs[k] = v
    return (attrs, encoding)
